@@ -10,7 +10,6 @@ import (
 	"log"
 	"os"
 	"testing"
-	"time"
 
 	"golang.org/x/tools/internal/lsp"
 	"golang.org/x/tools/internal/lsp/fake"
@@ -183,9 +182,12 @@ const test38878 = `
 -- go.mod --
 module foo
 
+go 1.12
 -- a.go --
 package x
 
+// import "fmt"
+
 func f() {}
 
 -- a_test.go --
@@ -198,28 +200,33 @@ func TestA(t *testing.T) {
 }
 `
 
-func TestRmTest38878Close(t *testing.T) {
+// Tests golang/go#38878: deleting a test file should clear its errors, and
+// not break the workspace.
+func TestDeleteTestVariant(t *testing.T) {
 	runner.Run(t, test38878, func(t *testing.T, env *Env) {
-		env.OpenFile("a_test.go")
-		env.Await(DiagnosticAt("a_test.go", 5, 3))
-		env.CloseBuffer("a_test.go")
+		env.Await(env.DiagnosticAtRegexp("a_test.go", `f\((3)\)`))
 		env.RemoveWorkspaceFile("a_test.go")
-		// diagnostics go away
 		env.Await(EmptyDiagnostics("a_test.go"))
+
+		// Make sure the test variant has been removed from the workspace by
+		// triggering a metadata load.
+		env.OpenFile("a.go")
+		env.RegexpReplace("a.go", `// import`, "import")
+		env.Await(env.DiagnosticAtRegexp("a.go", `"fmt"`))
 	})
 }
 
-func TestRmTest38878(t *testing.T) {
+// Tests golang/go#38878: deleting a test file on disk while it's still open
+// should not clear its errors.
+func TestDeleteTestVariant_DiskOnly(t *testing.T) {
 	log.SetFlags(log.Lshortfile)
 	runner.Run(t, test38878, func(t *testing.T, env *Env) {
 		env.OpenFile("a_test.go")
 		env.Await(DiagnosticAt("a_test.go", 5, 3))
 		env.Sandbox.Workdir.RemoveFile(context.Background(), "a_test.go")
-		// diagnostics remain after giving gopls a chance to do something
-		// (there is not yet a better way to decide gopls isn't going
-		// to do anything)
-		time.Sleep(time.Second)
-		env.Await(DiagnosticAt("a_test.go", 5, 3))
+		env.Await(OnceMet(
+			CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidChangeWatchedFiles), 1),
+			DiagnosticAt("a_test.go", 5, 3)))
 	})
 }
 
@@ -283,7 +290,7 @@ func Hello() {
 
 	t.Run("without workspace module", func(t *testing.T) {
 		withOptions(
-			WithModes(WithoutExperiments),
+			Modes(Singleton),
 		).run(t, noMod, func(t *testing.T, env *Env) {
 			env.Await(
 				env.DiagnosticAtRegexp("main.go", `"mod.com/bob"`),
@@ -352,6 +359,8 @@ func TestPackageChange_Issue38328(t *testing.T) {
 	const packageChange = `
 -- go.mod --
 module fake
+
+go 1.12
 -- a.go --
 package foo
 func main() {}
@@ -379,9 +388,10 @@ module mod.com
 
 go 1.12
 
-require (
-	foo.test v1.2.3
-)
+require foo.test v1.2.3
+-- go.sum --
+foo.test v1.2.3 h1:TMA+lyd1ck0TqjSFpNe4T6cf/K6TYkoHwOOcMBMjaEw=
+foo.test v1.2.3/go.mod h1:Ij3kyLIe5lzjycjh13NL8I2gX0quZuTdW0MnmlwGBL4=
 -- print.go --
 package lib
 
@@ -414,7 +424,7 @@ func TestResolveDiagnosticWithDownload(t *testing.T) {
 		// diagnostic for the wrong formatting type.
 		// TODO: we should be able to easily also match the diagnostic message.
 		env.Await(env.DiagnosticAtRegexp("print.go", "fmt.Printf"))
-	}, WithProxyFiles(testPackageWithRequireProxy))
+	}, ProxyFiles(testPackageWithRequireProxy))
 }
 
 func TestMissingDependency(t *testing.T) {
@@ -450,8 +460,13 @@ func _() {
 	fmt.Println("Hello World")
 }
 `
-	editorConfig := EditorConfig{Env: map[string]string{"GOPATH": ""}}
-	withOptions(editorConfig).run(t, files, func(t *testing.T, env *Env) {
+	withOptions(
+		EditorConfig{
+			Env: map[string]string{
+				"GOPATH":      "",
+				"GO111MODULE": "off",
+			},
+		}).run(t, files, func(t *testing.T, env *Env) {
 		env.OpenFile("main.go")
 		env.Await(env.DiagnosticAtRegexp("main.go", "fmt"))
 		env.SaveBuffer("main.go")
@@ -465,6 +480,7 @@ func TestEqualInEnv_Issue38669(t *testing.T) {
 -- go.mod --
 module mod.com
 
+go 1.12
 -- main.go --
 package main
 
@@ -488,6 +504,7 @@ func TestNoSuggestedFixesForGeneratedFiles_Issue38467(t *testing.T) {
 -- go.mod --
 module mod.com
 
+go 1.12
 -- main.go --
 package main
 
@@ -521,43 +538,26 @@ func _() {
 
 // Expect a module/GOPATH error if there is an error in the file at startup.
 // Tests golang/go#37279.
-func TestShowMessage_Issue37279(t *testing.T) {
+func TestShowCriticalError_Issue37279(t *testing.T) {
 	const noModule = `
 -- a.go --
 package foo
 
-func f() {
-	fmt.Printl()
-}
-`
-	runner.Run(t, noModule, func(t *testing.T, env *Env) {
-		env.OpenFile("a.go")
-		env.Await(env.DiagnosticAtRegexp("a.go", "fmt.Printl"), ShownMessage(""))
-	})
-}
-
-// Expect no module/GOPATH error if there is no error in the file.
-// Tests golang/go#37279.
-func TestNoShowMessage_Issue37279(t *testing.T) {
-	const noModule = `
--- a.go --
-package foo
+import "mod.com/hello"
 
 func f() {
+	hello.Goodbye()
 }
 `
 	runner.Run(t, noModule, func(t *testing.T, env *Env) {
 		env.OpenFile("a.go")
 		env.Await(
-			OnceMet(
-				CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidOpen), 1),
-				NoDiagnostics("a.go"),
-			),
-			NoShowMessage(),
+			OutstandingWork(lsp.WorkspaceLoadFailure, "outside of a module"),
+		)
+		env.RegexpReplace("a.go", `import "mod.com/hello"`, "")
+		env.Await(
+			NoOutstandingWork(),
 		)
-		// introduce an error, expect no Show Message
-		env.RegexpReplace("a.go", "func", "fun")
-		env.Await(env.DiagnosticAtRegexp("a.go", "fun"), NoShowMessage())
 	})
 }
 
@@ -655,7 +655,7 @@ func main() {
 }
 `
 	withOptions(
-		WithProxyFiles(ardanLabsProxy),
+		ProxyFiles(ardanLabsProxy),
 	).run(t, ardanLabs, func(t *testing.T, env *Env) {
 		// Expect a diagnostic with a suggested fix to add
 		// "github.com/ardanlabs/conf" to the go.mod file.
@@ -713,7 +713,7 @@ go 1.12
 -- main.go --
 `
 	withOptions(
-		WithProxyFiles(ardanLabsProxy),
+		ProxyFiles(ardanLabsProxy),
 	).run(t, emptyFile, func(t *testing.T, env *Env) {
 		env.OpenFile("main.go")
 		env.EditBuffer("main.go", fake.NewEdit(0, 0, 0, 0, `package main
@@ -728,7 +728,7 @@ func main() {
 		var d protocol.PublishDiagnosticsParams
 		env.Await(
 			OnceMet(
-				env.DiagnosticAtRegexp("main.go", `"github.com/ardanlabs/conf"`),
+				env.DiagnosticAtRegexpWithMessage("main.go", `"github.com/ardanlabs/conf"`, "your go.mod file"),
 				ReadDiagnostics("main.go", &d),
 			),
 		)
@@ -852,9 +852,13 @@ func TestHello(t *testing.T) {
 
 // Reproduce golang/go#40690.
 func TestCreateOnlyXTest(t *testing.T) {
+	testenv.NeedsGo1Point(t, 13)
+
 	const mod = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- foo/foo.go --
 package foo
 -- foo/bar_test.go --
@@ -885,6 +889,8 @@ func TestChangePackageName(t *testing.T) {
 	const mod = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- foo/foo.go --
 package foo
 -- foo/bar_test.go --
@@ -930,7 +936,7 @@ func TestIgnoredFiles(t *testing.T) {
 -- go.mod --
 module mod.com
 
-go 1.15
+go 1.12
 -- _foo/x.go --
 package x
 
@@ -1212,6 +1218,8 @@ func TestInvalidPackageName(t *testing.T) {
 	const pkgDefault = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- main.go --
 package default
 
@@ -1230,6 +1238,8 @@ func TestLimitWorkspaceScope(t *testing.T) {
 	const mod = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- a/main.go --
 package main
 
@@ -1241,13 +1251,13 @@ func main() {
 	var x int
 }
 `
-	withOptions(WithRootPath("a")).run(t, mod, func(t *testing.T, env *Env) {
+	withOptions(RootPath("a")).run(t, mod, func(t *testing.T, env *Env) {
 		env.OpenFile("a/main.go")
 		env.Await(
 			env.DiagnosticAtRegexp("main.go", "x"),
 		)
 	})
-	withOptions(WithRootPath("a"), LimitWorkspaceScope()).run(t, mod, func(t *testing.T, env *Env) {
+	withOptions(RootPath("a"), LimitWorkspaceScope()).run(t, mod, func(t *testing.T, env *Env) {
 		env.OpenFile("a/main.go")
 		env.Await(
 			NoDiagnostics("main.go"),
@@ -1259,6 +1269,8 @@ func TestStaticcheckDiagnostic(t *testing.T) {
 	const files = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- main.go --
 package main
 
@@ -1288,6 +1300,8 @@ func TestSecondaryDiagnostics(t *testing.T) {
 	const dir = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- main.go --
 package main
 func main() {
@@ -1365,6 +1379,7 @@ func TestEnableAllExperiments(t *testing.T) {
 -- go.mod --
 module mod.com
 
+go 1.12
 -- main.go --
 package main
 
@@ -1390,6 +1405,8 @@ func TestSwig(t *testing.T) {
 	const mod = `
 -- go.mod --
 module mod.com
+
+go 1.12
 -- pkg/simple/export_swig.go --
 package simple
 
@@ -1434,22 +1451,44 @@ func main() {
 // have no more complaints about it.
 // https://github.com/golang/go/issues/41061
 func TestRenamePackage(t *testing.T) {
-	t.Skip("Waiting for the fix that makes this pass: https://github.com/golang/go/issues/41061")
+	testenv.NeedsGo1Point(t, 16)
 
 	const contents = `
 -- go.mod --
 module mod.com
--- foo.go --
-package foo
--- foo_test.go --
-package foo_`
 
-	runner.Run(t, contents, func(t *testing.T, env *Env) {
-		env.OpenFile("foo_test.go")
-		env.RegexpReplace("foo_test.go", "foo_", "foo_test")
-		env.SaveBuffer("foo_test.go")
+go 1.12
+-- main.go --
+package main
+
+import "example.com/blah"
+
+func main() {
+	blah.Hello()
+}
+-- bob.go --
+package main
+-- foo/foo.go --
+package foo
+-- foo/foo_test.go --
+package foo_
+`
+
+	withOptions(
+		ProxyFiles(proxy),
+		InGOPATH(),
+	).run(t, contents, func(t *testing.T, env *Env) {
+		// Simulate typing character by character.
+		env.OpenFile("foo/foo_test.go")
+		env.Await(CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidOpen), 1))
+		env.RegexpReplace("foo/foo_test.go", "_", "_t")
+		env.Await(CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidChange), 1))
+		env.RegexpReplace("foo/foo_test.go", "_t", "_test")
+		env.Await(CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidChange), 2))
+
 		env.Await(
-			EmptyDiagnostics("foo_test.go"),
+			EmptyDiagnostics("foo/foo_test.go"),
+			NoOutstandingWork(),
 		)
 	})
 }
@@ -1470,14 +1509,14 @@ package main
 	run(t, pkg, func(t *testing.T, env *Env) {
 		env.OpenFile("go.mod")
 		env.Await(
-			OutstandingWork("Error loading workspace", "unknown directive"),
+			OutstandingWork(lsp.WorkspaceLoadFailure, "unknown directive"),
 		)
 		env.EditBuffer("go.mod", fake.NewEdit(0, 0, 3, 0, `module mod.com
 
 go 1.hello
 `))
 		env.Await(
-			OutstandingWork("Error loading workspace", "invalid go version"),
+			OutstandingWork(lsp.WorkspaceLoadFailure, "invalid go version"),
 		)
 		env.RegexpReplace("go.mod", "go 1.hello", "go 1.12")
 		env.Await(
@@ -1516,6 +1555,212 @@ func main() {
 	})
 }
 
+// Confirms that circular imports are tested and reported.
+func TestCircularImports(t *testing.T) {
+	const mod = `
+-- go.mod --
+module mod.com
+
+go 1.12
+-- self/self.go --
+package self
+
+import _ "mod.com/self"
+func Hello() {}
+-- double/a/a.go --
+package a
+
+import _ "mod.com/double/b"
+-- double/b/b.go --
+package b
+
+import _ "mod.com/double/a"
+-- triple/a/a.go --
+package a
+
+import _ "mod.com/triple/b"
+-- triple/b/b.go --
+package b
+
+import _ "mod.com/triple/c"
+-- triple/c/c.go --
+package c
+
+import _ "mod.com/triple/a"
+`
+	run(t, mod, func(t *testing.T, env *Env) {
+		env.Await(
+			env.DiagnosticAtRegexpWithMessage("self/self.go", `_ "mod.com/self"`, "import cycle not allowed"),
+			env.DiagnosticAtRegexpWithMessage("double/a/a.go", `_ "mod.com/double/b"`, "import cycle not allowed"),
+			env.DiagnosticAtRegexpWithMessage("triple/a/a.go", `_ "mod.com/triple/b"`, "import cycle not allowed"),
+		)
+	})
+}
+
+func TestBadImport(t *testing.T) {
+	testenv.NeedsGo1Point(t, 14)
+
+	const mod = `
+-- go.mod --
+module mod.com
+
+go 1.12
+-- main.go --
+package main
+
+import (
+	_ "nosuchpkg"
+)
+`
+	t.Run("module", func(t *testing.T) {
+		run(t, mod, func(t *testing.T, env *Env) {
+			env.Await(
+				env.DiagnosticAtRegexpWithMessage("main.go", `"nosuchpkg"`, `could not import nosuchpkg (no required module provides package "nosuchpkg"`),
+			)
+		})
+	})
+	t.Run("GOPATH", func(t *testing.T) {
+		withOptions(
+			InGOPATH(),
+			EditorConfig{
+				Env: map[string]string{"GO111MODULE": "off"},
+			},
+			Modes(Singleton),
+		).run(t, mod, func(t *testing.T, env *Env) {
+			env.Await(
+				env.DiagnosticAtRegexpWithMessage("main.go", `"nosuchpkg"`, `cannot find package "nosuchpkg" in any of`),
+			)
+		})
+	})
+}
+
+func TestMultipleModules_Warning(t *testing.T) {
+	const modules = `
+-- a/go.mod --
+module a.com
+
+go 1.12
+-- a/a.go --
+package a
+-- b/go.mod --
+module b.com
+
+go 1.12
+-- b/b.go --
+package b
+`
+	for _, go111module := range []string{"on", "auto"} {
+		t.Run("GO111MODULE="+go111module, func(t *testing.T) {
+			withOptions(
+				Modes(Singleton),
+				EditorConfig{
+					Env: map[string]string{
+						"GO111MODULE": go111module,
+					},
+				},
+			).run(t, modules, func(t *testing.T, env *Env) {
+				env.OpenFile("a/a.go")
+				env.OpenFile("b/go.mod")
+				env.Await(
+					env.DiagnosticAtRegexp("a/a.go", "package a"),
+					env.DiagnosticAtRegexp("b/go.mod", "module b.com"),
+					OutstandingWork(lsp.WorkspaceLoadFailure, "gopls requires a module at the root of your workspace."),
+				)
+			})
+		})
+	}
+
+	// Expect no warning if GO111MODULE=auto in a directory in GOPATH.
+	t.Run("GOPATH_GO111MODULE_auto", func(t *testing.T) {
+		withOptions(
+			Modes(Singleton),
+			EditorConfig{
+				Env: map[string]string{
+					"GO111MODULE": "auto",
+				},
+			},
+			InGOPATH(),
+		).run(t, modules, func(t *testing.T, env *Env) {
+			env.OpenFile("a/a.go")
+			env.Await(
+				OnceMet(
+					CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidOpen), 1),
+					NoDiagnostics("a/a.go"),
+				),
+				NoOutstandingWork(),
+			)
+		})
+	})
+}
+
+func TestNestedModules(t *testing.T) {
+	const proxy = `
+-- nested.com@v1.0.0/go.mod --
+module nested.com
+
+go 1.12
+-- nested.com@v1.0.0/hello/hello.go --
+package hello
+
+func Hello() {}
+`
+
+	const nested = `
+-- go.mod --
+module mod.com
+
+go 1.12
+
+require nested.com v1.0.0
+-- go.sum --
+nested.com v1.0.0 h1:I6spLE4CgFqMdBPc+wTV2asDO2QJ3tU0YAT+jkLeN1I=
+nested.com v1.0.0/go.mod h1:ly53UzXQgVjSlV7wicdBB4p8BxfytuGT1Xcyv0ReJfI=
+-- main.go --
+package main
+
+import "nested.com/hello"
+
+func main() {
+	hello.Hello()
+}
+-- nested/go.mod --
+module nested.com
+
+-- nested/hello/hello.go --
+package hello
+
+func Hello() {
+	helloHelper()
+}
+-- nested/hello/hello_helper.go --
+package hello
+
+func helloHelper() {}
+`
+	withOptions(
+		ProxyFiles(proxy),
+		Modes(Singleton),
+	).run(t, nested, func(t *testing.T, env *Env) {
+		// Expect a diagnostic in a nested module.
+		env.OpenFile("nested/hello/hello.go")
+		didOpen := CompletedWork(lsp.DiagnosticWorkTitle(lsp.FromDidOpen), 1)
+		env.Await(
+			OnceMet(
+				didOpen,
+				env.DiagnosticAtRegexp("nested/hello/hello.go", "helloHelper"),
+			),
+			OnceMet(
+				didOpen,
+				env.DiagnosticAtRegexpWithMessage("nested/hello/hello.go", "package hello", "nested module"),
+			),
+			OnceMet(
+				didOpen,
+				OutstandingWork(lsp.WorkspaceLoadFailure, "nested module"),
+			),
+		)
+	})
+}
+
 func TestAdHocPackagesReloading(t *testing.T) {
 	const nomod = `
 -- main.go --
