@@ -63,10 +63,10 @@ type EditorConfig struct {
 	Env        map[string]string
 	BuildFlags []string
 
-	// CodeLens is a map defining whether codelens are enabled, keyed by the
-	// codeLens command. CodeLens which are not present in this map are left in
+	// CodeLenses is a map defining whether codelens are enabled, keyed by the
+	// codeLens command. CodeLenses which are not present in this map are left in
 	// their default state.
-	CodeLens map[string]bool
+	CodeLenses map[string]bool
 
 	// SymbolMatcher is the config associated with the "symbolMatcher" gopls
 	// config option.
@@ -81,9 +81,9 @@ type EditorConfig struct {
 	// workspace folders nor a root URI.
 	WithoutWorkspaceFolders bool
 
-	// EditorRootPath specifies the root path of the workspace folder used when
+	// WorkspaceRoot specifies the root path of the workspace folder used when
 	// initializing gopls in the sandbox. If empty, the Workdir is used.
-	EditorRootPath string
+	WorkspaceRoot string
 
 	// EnableStaticcheck enables staticcheck analyzers.
 	EnableStaticcheck bool
@@ -95,6 +95,10 @@ type EditorConfig struct {
 	// Whether to send the current process ID, for testing data that is joined to
 	// the PID. This can only be set by one test.
 	SendPID bool
+
+	DirectoryFilters []string
+
+	VerboseOutput bool
 }
 
 // NewEditor Creates a new Editor.
@@ -121,7 +125,7 @@ func (e *Editor) Connect(ctx context.Context, conn jsonrpc2.Conn, hooks ClientHo
 		protocol.Handlers(
 			protocol.ClientHandler(e.client,
 				jsonrpc2.MethodNotFound)))
-	if err := e.initialize(ctx, e.Config.WithoutWorkspaceFolders, e.Config.EditorRootPath); err != nil {
+	if err := e.initialize(ctx, e.Config.WithoutWorkspaceFolders, e.Config.WorkspaceRoot); err != nil {
 		return nil, err
 	}
 	e.sandbox.Workdir.AddWatcher(e.onFileChanges)
@@ -195,9 +199,11 @@ func (e *Editor) configuration() map[string]interface{} {
 	if e.Config.BuildFlags != nil {
 		config["buildFlags"] = e.Config.BuildFlags
 	}
-
-	if e.Config.CodeLens != nil {
-		config["codelens"] = e.Config.CodeLens
+	if e.Config.DirectoryFilters != nil {
+		config["directoryFilters"] = e.Config.DirectoryFilters
+	}
+	if e.Config.CodeLenses != nil {
+		config["codelenses"] = e.Config.CodeLenses
 	}
 	if e.Config.SymbolMatcher != nil {
 		config["symbolMatcher"] = *e.Config.SymbolMatcher
@@ -212,9 +218,13 @@ func (e *Editor) configuration() map[string]interface{} {
 		config["allExperiments"] = true
 	}
 
-	// TODO(rFindley): uncomment this if/when diagnostics delay is on by
-	// default... and probably change to the new settings name.
-	// config["experimentalDiagnosticsDelay"] = "10ms"
+	if e.Config.VerboseOutput {
+		config["verboseOutput"] = true
+	}
+
+	// TODO(rFindley): change to the new settings name once it is no longer
+	// designated experimental.
+	config["experimentalDiagnosticsDelay"] = "10ms"
 
 	// ExperimentalWorkspaceModule is only set as a mode, not a configuration.
 	return config
@@ -391,9 +401,9 @@ func (e *Editor) SaveBuffer(ctx context.Context, path string) error {
 
 func (e *Editor) SaveBufferWithoutActions(ctx context.Context, path string) error {
 	e.mu.Lock()
+	defer e.mu.Unlock()
 	buf, ok := e.buffers[path]
 	if !ok {
-		e.mu.Unlock()
 		return fmt.Errorf(fmt.Sprintf("unknown buffer: %q", path))
 	}
 	content := buf.text()
@@ -402,7 +412,6 @@ func (e *Editor) SaveBufferWithoutActions(ctx context.Context, path string) erro
 	if ok {
 		includeText = syncOptions.Save.IncludeText
 	}
-	e.mu.Unlock()
 
 	docID := e.textDocumentIdentifier(buf.path)
 	if e.Server != nil {
@@ -417,10 +426,8 @@ func (e *Editor) SaveBufferWithoutActions(ctx context.Context, path string) erro
 		return errors.Errorf("writing %q: %w", path, err)
 	}
 
-	e.mu.Lock()
 	buf.dirty = false
 	e.buffers[path] = buf
-	e.mu.Unlock()
 
 	if e.Server != nil {
 		params := &protocol.DidSaveTextDocumentParams{
@@ -806,6 +813,9 @@ func (e *Editor) FormatBuffer(ctx context.Context, path string) error {
 		return fmt.Errorf("before receipt of formatting edits, buffer version changed from %d to %d", version, versionAfter)
 	}
 	edits := convertEdits(resp)
+	if len(edits) == 0 {
+		return nil
+	}
 	return e.editBufferLocked(ctx, path, edits)
 }
 
