@@ -12,9 +12,9 @@ import (
 	"os/exec"
 	"path/filepath"
 	"sort"
+	"strings"
 	"testing"
 
-	"golang.org/x/tools/go/packages/packagestest"
 	"golang.org/x/tools/internal/lsp/cache"
 	"golang.org/x/tools/internal/lsp/diff"
 	"golang.org/x/tools/internal/lsp/diff/myers"
@@ -31,29 +31,27 @@ func TestMain(m *testing.M) {
 }
 
 func TestLSP(t *testing.T) {
-	packagestest.TestAll(t, testLSP)
+	tests.RunTests(t, "testdata", true, testLSP)
 }
 
 type runner struct {
 	server      *Server
 	data        *tests.Data
-	diagnostics map[span.URI]map[string]*source.Diagnostic
+	diagnostics map[span.URI][]*source.Diagnostic
 	ctx         context.Context
+	normalizers []tests.Normalizer
 }
 
-func testLSP(t *testing.T, exporter packagestest.Exporter) {
+func testLSP(t *testing.T, datum *tests.Data) {
 	ctx := tests.Context(t)
-	datum := tests.Load(t, exporter, "testdata")
-	defer datum.Exported.Cleanup()
 
 	cache := cache.New(ctx, nil)
 	session := cache.NewSession(ctx)
 	options := source.DefaultOptions().Clone()
 	tests.DefaultOptions(options)
 	session.SetOptions(options)
-	options.Env = datum.Config.Env
-	view, _, release, err := session.NewView(ctx, datum.Config.Dir, span.URIFromPath(datum.Config.Dir), options)
-	release()
+	options.SetEnvSlice(datum.Config.Env)
+	view, snapshot, release, err := session.NewView(ctx, datum.Config.Dir, span.URIFromPath(datum.Config.Dir), "", options)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -66,7 +64,8 @@ func testLSP(t *testing.T, exporter packagestest.Exporter) {
 	view.SetOptions(ctx, options)
 
 	// Only run the -modfile specific tests in module mode with Go 1.14 or above.
-	datum.ModfileFlagAvailable = view.ModFile() != "" && testenv.Go1Point() >= 14
+	datum.ModfileFlagAvailable = len(snapshot.ModFiles()) > 0 && testenv.Go1Point() >= 14
+	release()
 
 	var modifications []source.FileModification
 	for filename, content := range datum.Config.Overlay {
@@ -86,12 +85,23 @@ func testLSP(t *testing.T, exporter packagestest.Exporter) {
 		t.Fatal(err)
 	}
 	r := &runner{
-		server: NewServer(session, nil),
-		data:   datum,
-		ctx:    ctx,
+		server:      NewServer(session, testClient{}),
+		data:        datum,
+		ctx:         ctx,
+		normalizers: tests.CollectNormalizers(datum.Exported),
 	}
 	tests.Run(t, r, datum)
+}
 
+// testClient stubs any client functions that may be called by LSP functions.
+type testClient struct {
+	protocol.Client
+}
+
+// Trivially implement PublishDiagnostics so that we can call
+// server.publishReports below to de-dup sent diagnostics.
+func (c testClient) PublishDiagnostics(context.Context, *protocol.PublishDiagnosticsParams) error {
+	return nil
 }
 
 func (r *runner) CallHierarchy(t *testing.T, spn span.Span, expectedCalls *tests.CallHierarchyResult) {
@@ -173,21 +183,11 @@ func (r *runner) CodeLens(t *testing.T, uri span.URI, want []protocol.CodeLens)
 
 func (r *runner) Diagnostics(t *testing.T, uri span.URI, want []*source.Diagnostic) {
 	// Get the diagnostics for this view if we have not done it before.
-	if r.diagnostics == nil {
-		r.diagnostics = make(map[span.URI]map[string]*source.Diagnostic)
-		v := r.server.session.View(r.data.Config.Dir)
-		// Always run diagnostics with analysis.
-		snapshot, release := v.Snapshot(r.ctx)
-		defer release()
-		reports, _ := r.server.diagnose(r.ctx, snapshot, true)
-		for key, diags := range reports {
-			r.diagnostics[key.id.URI] = diags
-		}
-	}
-	var got []*source.Diagnostic
-	for _, d := range r.diagnostics[uri] {
-		got = append(got, d)
-	}
+	v := r.server.session.View(r.data.Config.Dir)
+	r.collectDiagnostics(v)
+	d := r.diagnostics[uri]
+	got := make([]*source.Diagnostic, len(d))
+	copy(got, d)
 	// A special case to test that there are no diagnostics for a file.
 	if len(want) == 1 && want[0].Source == "no_diagnostics" {
 		if len(got) != 0 {
@@ -393,6 +393,39 @@ func (r *runner) Format(t *testing.T, spn span.Span) {
 	}
 }
 
+func (r *runner) SemanticTokens(t *testing.T, spn span.Span) {
+	uri := spn.URI()
+	filename := uri.Filename()
+	// this is called solely for coverage in semantic.go
+	_, err := r.server.semanticTokensFull(r.ctx, &protocol.SemanticTokensParams{
+		TextDocument: protocol.TextDocumentIdentifier{
+			URI: protocol.URIFromSpanURI(uri),
+		},
+	})
+	if err != nil {
+		t.Errorf("%v for %s", err, filename)
+	}
+	_, err = r.server.semanticTokensRange(r.ctx, &protocol.SemanticTokensRangeParams{
+		TextDocument: protocol.TextDocumentIdentifier{
+			URI: protocol.URIFromSpanURI(uri),
+		},
+		// any legal range. Just to exercise the call.
+		Range: protocol.Range{
+			Start: protocol.Position{
+				Line:      0,
+				Character: 0,
+			},
+			End: protocol.Position{
+				Line:      2,
+				Character: 0,
+			},
+		},
+	})
+	if err != nil {
+		t.Errorf("%v for Range %s", err, filename)
+	}
+}
+
 func (r *runner) Import(t *testing.T, spn span.Span) {
 	uri := spn.URI()
 	filename := uri.Filename()
@@ -425,7 +458,7 @@ func (r *runner) Import(t *testing.T, spn span.Span) {
 	}
 }
 
-func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string) {
+func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string, expectedActions int) {
 	uri := spn.URI()
 	view, err := r.server.session.ViewOf(uri)
 	if err != nil {
@@ -435,7 +468,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string)
 	snapshot, release := view.Snapshot(r.ctx)
 	defer release()
 
-	fh, err := snapshot.GetFile(r.ctx, uri)
+	fh, err := snapshot.GetVersionedFile(r.ctx, uri)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -448,16 +481,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string)
 		t.Fatal(err)
 	}
 	// Get the diagnostics for this view if we have not done it before.
-	if r.diagnostics == nil {
-		r.diagnostics = make(map[span.URI]map[string]*source.Diagnostic)
-		// Always run diagnostics with analysis.
-		snapshot, release := view.Snapshot(r.ctx)
-		defer release()
-		reports, _ := r.server.diagnose(r.ctx, snapshot, true)
-		for key, diags := range reports {
-			r.diagnostics[key.id.URI] = diags
-		}
-	}
+	r.collectDiagnostics(view)
 	var diagnostics []protocol.Diagnostic
 	for _, d := range r.diagnostics[uri] {
 		// Compare the start positions rather than the entire range because
@@ -485,10 +509,13 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string)
 	if err != nil {
 		t.Fatalf("CodeAction %s failed: %v", spn, err)
 	}
-	if len(actions) != 1 {
+	if len(actions) != expectedActions {
 		// Hack: We assume that we only get one code action per range.
-		// TODO(rstambler): Support multiple code actions per test.
-		t.Fatalf("unexpected number of code actions, want 1, got %v", len(actions))
+		var cmds []string
+		for _, a := range actions {
+			cmds = append(cmds, fmt.Sprintf("%s (%s)", a.Command.Command, a.Title))
+		}
+		t.Fatalf("unexpected number of code actions, want %d, got %d: %v", expectedActions, len(actions), cmds)
 	}
 	action := actions[0]
 	var match bool
@@ -505,7 +532,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string)
 	if cmd := action.Command; cmd != nil {
 		edits, err := commandToEdits(r.ctx, snapshot, fh, rng, action.Command.Command)
 		if err != nil {
-			t.Fatal(err)
+			t.Fatalf("error converting command %q to edits: %v", action.Command.Command, err)
 		}
 		res, err = applyTextDocumentEdits(r, edits)
 		if err != nil {
@@ -530,7 +557,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []string)
 func commandToEdits(ctx context.Context, snapshot source.Snapshot, fh source.VersionedFileHandle, rng protocol.Range, cmd string) ([]protocol.TextDocumentEdit, error) {
 	var command *source.Command
 	for _, c := range source.Commands {
-		if c.Name == cmd {
+		if c.ID() == cmd {
 			command = c
 			break
 		}
@@ -539,9 +566,13 @@ func commandToEdits(ctx context.Context, snapshot source.Snapshot, fh source.Ver
 		return nil, fmt.Errorf("no known command for %s", cmd)
 	}
 	if !command.Applies(ctx, snapshot, fh, rng) {
-		return nil, fmt.Errorf("cannot apply %v", command.Name)
+		return nil, fmt.Errorf("cannot apply %v", command.ID())
 	}
-	return command.SuggestedFix(ctx, snapshot, fh, rng)
+	edits, err := command.SuggestedFix(ctx, snapshot, fh, rng)
+	if err != nil {
+		return nil, fmt.Errorf("error calling command.SuggestedFix: %v", err)
+	}
+	return edits, nil
 }
 
 func (r *runner) FunctionExtraction(t *testing.T, start span.Span, end span.Span) {
@@ -554,7 +585,7 @@ func (r *runner) FunctionExtraction(t *testing.T, start span.Span, end span.Span
 	snapshot, release := view.Snapshot(r.ctx)
 	defer release()
 
-	fh, err := snapshot.GetFile(r.ctx, uri)
+	fh, err := snapshot.GetVersionedFile(r.ctx, uri)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -991,21 +1022,15 @@ func (r *runner) Symbols(t *testing.T, uri span.URI, expectedSymbols []protocol.
 	}
 }
 
-func (r *runner) WorkspaceSymbols(t *testing.T, query string, expectedSymbols []protocol.SymbolInformation, dirs map[string]struct{}) {
-	r.callWorkspaceSymbols(t, query, source.SymbolCaseInsensitive, dirs, expectedSymbols)
+func (r *runner) WorkspaceSymbols(t *testing.T, uri span.URI, query string, typ tests.WorkspaceSymbolsTestType) {
+	r.callWorkspaceSymbols(t, uri, query, typ)
 }
 
-func (r *runner) FuzzyWorkspaceSymbols(t *testing.T, query string, expectedSymbols []protocol.SymbolInformation, dirs map[string]struct{}) {
-	r.callWorkspaceSymbols(t, query, source.SymbolFuzzy, dirs, expectedSymbols)
-}
-
-func (r *runner) CaseSensitiveWorkspaceSymbols(t *testing.T, query string, expectedSymbols []protocol.SymbolInformation, dirs map[string]struct{}) {
-	r.callWorkspaceSymbols(t, query, source.SymbolCaseSensitive, dirs, expectedSymbols)
-}
-
-func (r *runner) callWorkspaceSymbols(t *testing.T, query string, matcher source.SymbolMatcher, dirs map[string]struct{}, expectedSymbols []protocol.SymbolInformation) {
+func (r *runner) callWorkspaceSymbols(t *testing.T, uri span.URI, query string, typ tests.WorkspaceSymbolsTestType) {
 	t.Helper()
 
+	matcher := tests.WorkspaceSymbolsTestTypeToMatcher(typ)
+
 	original := r.server.session.Options()
 	modified := original
 	modified.SymbolMatcher = matcher
@@ -1015,12 +1040,19 @@ func (r *runner) callWorkspaceSymbols(t *testing.T, query string, matcher source
 	params := &protocol.WorkspaceSymbolParams{
 		Query: query,
 	}
-	got, err := r.server.Symbol(r.ctx, params)
+	gotSymbols, err := r.server.Symbol(r.ctx, params)
 	if err != nil {
 		t.Fatal(err)
 	}
-	got = tests.FilterWorkspaceSymbols(got, dirs)
-	if diff := tests.DiffWorkspaceSymbols(expectedSymbols, got); diff != "" {
+	got, err := tests.WorkspaceSymbolsString(r.ctx, r.data, uri, gotSymbols)
+	if err != nil {
+		t.Fatal(err)
+	}
+	got = filepath.ToSlash(tests.Normalize(got, r.normalizers))
+	want := string(r.data.Golden(fmt.Sprintf("workspace_symbol-%s-%s", strings.ToLower(string(matcher)), query), uri.Filename(), func() ([]byte, error) {
+		return []byte(got), nil
+	}))
+	if diff := tests.Diff(want, got); diff != "" {
 		t.Error(diff)
 	}
 }
@@ -1126,3 +1158,32 @@ func TestBytesOffset(t *testing.T) {
 		}
 	}
 }
+
+func (r *runner) collectDiagnostics(view source.View) {
+	if r.diagnostics != nil {
+		return
+	}
+	r.diagnostics = make(map[span.URI][]*source.Diagnostic)
+
+	snapshot, release := view.Snapshot(r.ctx)
+	defer release()
+
+	// Always run diagnostics with analysis.
+	r.server.diagnose(r.ctx, snapshot, true)
+	for uri, reports := range r.server.diagnostics {
+		var diagnostics []*source.Diagnostic
+		for _, report := range reports.reports {
+			for _, d := range report.diags {
+				diagnostics = append(diagnostics, &source.Diagnostic{
+					Range:    d.Range,
+					Message:  d.Message,
+					Related:  d.Related,
+					Severity: d.Severity,
+					Source:   d.Source,
+					Tags:     d.Tags,
+				})
+			}
+			r.diagnostics[uri] = diagnostics
+		}
+	}
+}
