@@ -7,6 +7,7 @@ package source
 import (
 	"context"
 	"fmt"
+	"path/filepath"
 	"regexp"
 	"strings"
 	"sync"
@@ -24,6 +25,7 @@ import (
 	"golang.org/x/tools/go/analysis/passes/copylock"
 	"golang.org/x/tools/go/analysis/passes/deepequalerrors"
 	"golang.org/x/tools/go/analysis/passes/errorsas"
+	"golang.org/x/tools/go/analysis/passes/fieldalignment"
 	"golang.org/x/tools/go/analysis/passes/httpresponse"
 	"golang.org/x/tools/go/analysis/passes/ifaceassert"
 	"golang.org/x/tools/go/analysis/passes/loopclosure"
@@ -101,14 +103,7 @@ func DefaultOptions() *Options {
 			UserOptions: UserOptions{
 				HoverKind:  FullDocumentation,
 				LinkTarget: "pkg.go.dev",
-			},
-			DebuggingOptions: DebuggingOptions{
-				CompletionBudget: 100 * time.Millisecond,
-			},
-			ExperimentalOptions: ExperimentalOptions{
-				TempModfile:             true,
-				ExpandWorkspaceToModule: true,
-				Codelens: map[string]bool{
+				Codelenses: map[string]bool{
 					CommandGenerate.Name:          true,
 					CommandRegenerateCgo.Name:     true,
 					CommandTidy.Name:              true,
@@ -116,18 +111,26 @@ func DefaultOptions() *Options {
 					CommandUpgradeDependency.Name: true,
 					CommandVendor.Name:            true,
 				},
-				LinksInHover:                true,
-				CompleteUnimported:          true,
-				CompletionDocumentation:     true,
-				DeepCompletion:              true,
-				ImportShortcut:              Both,
-				Matcher:                     Fuzzy,
-				SymbolMatcher:               SymbolFuzzy,
-				SymbolStyle:                 PackageQualifiedSymbols,
-				ExperimentalPackageCacheKey: true,
+				LinksInHover:   true,
+				ImportShortcut: Both,
+				Matcher:        Fuzzy,
+				SymbolMatcher:  SymbolFuzzy,
+				SymbolStyle:    DynamicSymbols,
+			},
+			DebuggingOptions: DebuggingOptions{
+				CompletionBudget: 100 * time.Millisecond,
+			},
+			ExperimentalOptions: ExperimentalOptions{
+				ExpandWorkspaceToModule:      true,
+				ExperimentalPackageCacheKey:  true,
+				ExperimentalDiagnosticsDelay: 250 * time.Millisecond,
 			},
 			InternalOptions: InternalOptions{
-				LiteralCompletions: true,
+				LiteralCompletions:      true,
+				TempModfile:             true,
+				CompleteUnimported:      true,
+				CompletionDocumentation: true,
+				DeepCompletion:          true,
 			},
 			Hooks: Hooks{
 				ComputeEdits:         myers.ComputeEdits,
@@ -209,6 +212,75 @@ type UserOptions struct {
 
 	// Gofumpt indicates if we should run gofumpt formatting.
 	Gofumpt bool
+
+	// Analyses specify analyses that the user would like to enable or disable.
+	// A map of the names of analysis passes that should be enabled/disabled.
+	// A full list of analyzers that gopls uses can be found [here](analyzers.md)
+	//
+	// Example Usage:
+	// ```json5
+	// ...
+	// "analyses": {
+	//   "unreachable": false, // Disable the unreachable analyzer.
+	//   "unusedparams": true  // Enable the unusedparams analyzer.
+	// }
+	// ...
+	// ```
+	Analyses map[string]bool
+
+	// Codelenses overrides the enabled/disabled state of code lenses. See the "Code Lenses"
+	// section of settings.md for the list of supported lenses.
+	//
+	// Example Usage:
+	// ```json5
+	// "gopls": {
+	// ...
+	//   "codelens": {
+	//     "generate": false,  // Don't show the `go generate` lens.
+	//     "gc_details": true  // Show a code lens toggling the display of gc's choices.
+	//   }
+	// ...
+	// }
+	// ```
+	Codelenses map[string]bool
+
+	// LinksInHover toggles the presence of links to documentation in hover.
+	LinksInHover bool
+
+	// ImportShortcut specifies whether import statements should link to
+	// documentation or go to definitions.
+	ImportShortcut ImportShortcut
+
+	// Matcher sets the algorithm that is used when calculating completion candidates.
+	Matcher Matcher
+
+	// SymbolMatcher sets the algorithm that is used when finding workspace symbols.
+	SymbolMatcher SymbolMatcher
+
+	// SymbolStyle controls how symbols are qualified in symbol responses.
+	//
+	// Example Usage:
+	// ```json5
+	// "gopls": {
+	// ...
+	//   "symbolStyle": "dynamic",
+	// ...
+	// }
+	// ```
+	SymbolStyle SymbolStyle
+
+	// DirectoryFilters can be used to exclude unwanted directories from the
+	// workspace. By default, all directories are included. Filters are an
+	// operator, `+` to include and `-` to exclude, followed by a path prefix
+	// relative to the workspace folder. They are evaluated in order, and
+	// the last filter that applies to a path controls whether it is included.
+	// The path prefix can be empty, so an initial `-` excludes everything.
+	//
+	// Examples:
+	// Exclude node_modules: `-node_modules`
+	// Include only project_a: `-` (exclude everything), `+project_a`
+	// Include only project_a, but not node_modules inside it: `-`, `+project_a`, `-project_a/node_modules`
+	DirectoryFilters []string
 }
 
 // EnvSlice returns Env as a slice of k=v strings.
@@ -249,67 +321,6 @@ type Hooks struct {
 // development. WARNING: This configuration will be changed in the future. It
 // only exists while these features are under development.
 type ExperimentalOptions struct {
-	// Analyses specify analyses that the user would like to enable or disable.
-	// A map of the names of analysis passes that should be enabled/disabled.
-	// A full list of analyzers that gopls uses can be found [here](analyzers.md)
-	//
-	// Example Usage:
-	// ```json5
-	// ...
-	// "analyses": {
-	//   "unreachable": false, // Disable the unreachable analyzer.
-	//   "unusedparams": true  // Enable the unusedparams analyzer.
-	// }
-	// ...
-	// ```
-	Analyses map[string]bool
-
-	// Codelens overrides the enabled/disabled state of code lenses. See the "Code Lenses"
-	// section of settings.md for the list of supported lenses.
-	//
-	// Example Usage:
-	// ```json5
-	// "gopls": {
-	// ...
-	//   "codelens": {
-	//     "generate": false,  // Don't show the `go generate` lens.
-	//     "gc_details": true  // Show a code lens toggling the display of gc's choices.
-	//   }
-	// ...
-	// }
-	// ```
-	Codelens map[string]bool
-
-	// CompletionDocumentation enables documentation with completion results.
-	CompletionDocumentation bool
-
-	// CompleteUnimported enables completion for packages that you do not currently import.
-	CompleteUnimported bool
-
-	// DeepCompletion enables the ability to return completions from deep inside relevant entities, rather than just the locally accessible ones.
-	//
-	// Consider this example:
-	//
-	// ```go
-	// package main
-	//
-	// import "fmt"
-	//
-	// type wrapString struct {
-	//     str string
-	// }
-	//
-	// func main() {
-	//     x := wrapString{"hello world"}
-	//     fmt.Printf(<>)
-	// }
-	// ```
-	//
-	// At the location of the `<>` in this program, deep completion would suggest the result `x.str`.
-	DeepCompletion bool
-
-	// Matcher sets the algorithm that is used when calculating completion candidates.
-	Matcher Matcher
 
 	// Annotations suppress various kinds of optimization diagnostics
 	// that would be reported by the gc_details command.
@@ -322,35 +333,6 @@ type ExperimentalOptions struct {
 	// Staticcheck enables additional analyses from staticcheck.io.
 	Staticcheck bool
 
-	// SymbolMatcher sets the algorithm that is used when finding workspace symbols.
-	SymbolMatcher SymbolMatcher
-
-	// SymbolStyle controls how symbols are qualified in symbol responses.
-	//
-	// Example Usage:
-	// ```json5
-	// "gopls": {
-	// ...
-	//   "symbolStyle": "dynamic",
-	// ...
-	// }
-	// ```
-	SymbolStyle SymbolStyle
-
-	// LinksInHover toggles the presence of links to documentation in hover.
-	LinksInHover bool
-
-	// TempModfile controls the use of the -modfile flag in Go 1.14.
-	TempModfile bool
-
-	// ImportShortcut specifies whether import statements should link to
-	// documentation or go to definitions.
-	ImportShortcut ImportShortcut
-
-	// VerboseWorkDoneProgress controls whether the LSP server should send
-	// progress reports for all work done outside the scope of an RPC.
-	VerboseWorkDoneProgress bool
-
 	// SemanticTokens controls whether the LSP server will send
 	// semantic tokens to the client.
 	SemanticTokens bool
@@ -383,6 +365,15 @@ type ExperimentalOptions struct {
 	// by an experiment because caching behavior is subtle and difficult to
 	// comprehensively test.
 	ExperimentalPackageCacheKey bool
+
+	// AllowModfileModifications disables -mod=readonly, allowing imports from
+	// out-of-scope modules. This option will eventually be removed.
+	AllowModfileModifications bool
+
+	// AllowImplicitNetworkAccess disables GOPROXY=off, allowing implicit module
+	// downloads rather than requiring user action. This option will eventually
+	// be removed.
+	AllowImplicitNetworkAccess bool
 }
 
 // DebuggingOptions should not affect the logical execution of Gopls, but may
@@ -399,13 +390,56 @@ type DebuggingOptions struct {
 	CompletionBudget time.Duration
 }
 
-// InternalOptions contains settings that are not exposed to the user for various
-// reasons, e.g. settings used by tests.
+// InternalOptions contains settings that are not intended for use by the
+// average user. These may be settings used by tests or outdated settings that
+// will soon be deprecated. Some of these settings may not even be configurable
+// by the user.
 type InternalOptions struct {
 	// LiteralCompletions controls whether literal candidates such as
 	// "&someStruct{}" are offered. Tests disable this flag to simplify
 	// their expected values.
 	LiteralCompletions bool
+
+	// VerboseWorkDoneProgress controls whether the LSP server should send
+	// progress reports for all work done outside the scope of an RPC.
+	// Used by the regression tests.
+	VerboseWorkDoneProgress bool
+
+	// The following options were previously available to users, but they
+	// really shouldn't be configured by anyone other than "power users".
+
+	// CompletionDocumentation enables documentation with completion results.
+	CompletionDocumentation bool
+
+	// CompleteUnimported enables completion for packages that you do not
+	// currently import.
+	CompleteUnimported bool
+
+	// DeepCompletion enables the ability to return completions from deep
+	// inside relevant entities, rather than just the locally accessible ones.
+	//
+	// Consider this example:
+	//
+	// ```go
+	// package main
+	//
+	// import "fmt"
+	//
+	// type wrapString struct {
+	//     str string
+	// }
+	//
+	// func main() {
+	//     x := wrapString{"hello world"}
+	//     fmt.Printf(<>)
+	// }
+	// ```
+	//
+	// At the location of the `<>` in this program, deep completion would suggest the result `x.str`.
+	DeepCompletion bool
+
+	// TempModfile controls the use of the -modfile flag in Go 1.14.
+	TempModfile bool
 }
 
 type ImportShortcut string
@@ -577,7 +611,7 @@ func (o *Options) Clone() *Options {
 	}
 	result.Analyses = copyStringMap(o.Analyses)
 	result.Annotations = copyStringMap(o.Annotations)
-	result.Codelens = copyStringMap(o.Codelens)
+	result.Codelenses = copyStringMap(o.Codelenses)
 
 	copySlice := func(src []string) []string {
 		dst := make([]string, len(src))
@@ -586,6 +620,7 @@ func (o *Options) Clone() *Options {
 	}
 	result.SetEnvSlice(o.EnvSlice())
 	result.BuildFlags = copySlice(o.BuildFlags)
+	result.DirectoryFilters = copySlice(o.DirectoryFilters)
 
 	copyAnalyzerMap := func(src map[string]Analyzer) map[string]Analyzer {
 		dst := make(map[string]Analyzer)
@@ -606,17 +641,15 @@ func (o *Options) AddStaticcheckAnalyzer(a *analysis.Analyzer) {
 }
 
 // enableAllExperiments turns on all of the experimental "off-by-default"
-// features offered by gopls.
-// Any experimental features specified in maps should be enabled in
-// enableAllExperimentMaps.
+// features offered by gopls. Any experimental features specified in maps
+// should be enabled in enableAllExperimentMaps.
 func (o *Options) enableAllExperiments() {
-	o.ExperimentalDiagnosticsDelay = 200 * time.Millisecond
-	o.SymbolStyle = DynamicSymbols
+	// There are currently no experimental features in development.
 }
 
 func (o *Options) enableAllExperimentMaps() {
-	if _, ok := o.Codelens[CommandToggleDetails.Name]; !ok {
-		o.Codelens[CommandToggleDetails.Name] = true
+	if _, ok := o.Codelenses[CommandToggleDetails.Name]; !ok {
+		o.Codelenses[CommandToggleDetails.Name] = true
 	}
 	if _, ok := o.Analyses[unusedparams.Analyzer.Name]; !ok {
 		o.Analyses[unusedparams.Analyzer.Name] = true
@@ -647,7 +680,22 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 			flags = append(flags, fmt.Sprintf("%s", flag))
 		}
 		o.BuildFlags = flags
-
+	case "directoryFilters":
+		ifilters, ok := value.([]interface{})
+		if !ok {
+			result.errorf("invalid type %T, expect list", value)
+			break
+		}
+		var filters []string
+		for _, ifilter := range ifilters {
+			filter := fmt.Sprint(ifilter)
+			if filter[0] != '+' && filter[0] != '-' {
+				result.errorf("invalid filter %q, must start with + or -", filter)
+				return result
+			}
+			filters = append(filters, filepath.FromSlash(filter))
+		}
+		o.DirectoryFilters = filters
 	case "completionDocumentation":
 		result.setBool(&o.CompletionDocumentation)
 	case "usePlaceholders":
@@ -722,18 +770,25 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 			}
 		}
 
-	case "codelens":
+	case "codelenses", "codelens":
 		var lensOverrides map[string]bool
 		result.setBoolMap(&lensOverrides)
 		if result.Error == nil {
-			if o.Codelens == nil {
-				o.Codelens = make(map[string]bool)
+			if o.Codelenses == nil {
+				o.Codelenses = make(map[string]bool)
 			}
 			for lens, enabled := range lensOverrides {
-				o.Codelens[lens] = enabled
+				o.Codelenses[lens] = enabled
 			}
 		}
 
+		// codelens is deprecated, but still works for now.
+		// TODO(rstambler): Remove this for the gopls/v0.7.0 release.
+		if name == "codelens" {
+			result.State = OptionDeprecated
+			result.Replacement = "codelenses"
+		}
+
 	case "staticcheck":
 		result.setBool(&o.Staticcheck)
 
@@ -767,6 +822,12 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "experimentalPackageCacheKey":
 		result.setBool(&o.ExperimentalPackageCacheKey)
 
+	case "allowModfileModifications":
+		result.setBool(&o.AllowModfileModifications)
+
+	case "allowImplicitNetworkAccess":
+		result.setBool(&o.AllowImplicitNetworkAccess)
+
 	case "allExperiments":
 		// This setting should be handled before all of the other options are
 		// processed, so do nothing here.
@@ -992,6 +1053,7 @@ func defaultAnalyzers() map[string]Analyzer {
 		// Non-vet analyzers:
 		atomicalign.Analyzer.Name:      {Analyzer: atomicalign.Analyzer, Enabled: true},
 		deepequalerrors.Analyzer.Name:  {Analyzer: deepequalerrors.Analyzer, Enabled: true},
+		fieldalignment.Analyzer.Name:   {Analyzer: fieldalignment.Analyzer, Enabled: false},
 		sortslice.Analyzer.Name:        {Analyzer: sortslice.Analyzer, Enabled: true},
 		testinggoroutine.Analyzer.Name: {Analyzer: testinggoroutine.Analyzer, Enabled: true},
 		unusedparams.Analyzer.Name:     {Analyzer: unusedparams.Analyzer, Enabled: false},
